function [gC,HC,gE,HE,delta,output] = dcritstep(x,fun,fx,delta,gamma1,epsC,epsE)
%
% A decoupled criticality step for building two trust-region models.
%
% [gC,HC,gE,HE] = dcritstep(X0,FUN) computes two quadratic models 
% of linear terms (gC,gE) and quadratic terms (HC,HE), respectively.
%
% Inputs:
%		x: current point
%		fun: function handle
%		fx: value of fun at x
%		delta: current trust-region radius
%		gamma1: decrease parameter
%		epsC: first-order tolerance
%		epsE: second-order tolerance
%
% Outputs:
%		gC: linear part for the first-order model
%		HC: quadratic part for the first-order model
%		gE: linear part for the second-order model
%		HE: quadratic part for the second-order model
%		delta: final value for the trust-region radius
%		output: structure containing various information about the algorithm
%			nevals: number of calls to the function fun
%			its: iterations considered to have been performed
%			flagC: termination flag for building the first-order model
%			flagE: termination flag for building the second-order model
%
% Source:
%		A decoupled first/second-order steps technique for nonconvex nonlinear
%		unconstrained optimization with improved complexity bounds
%		S. Gratton, C. W. Royer and L. N. Vicente
%		Preprint 17-21, Dept. Mathematics, Univ. Coimbra, 2017.
%		(Revised version, January 2018)
%
% Implementation: C. W. Royer, 2018/07/22
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
	r=2;
	hepsC=(gamma1^r)*epsC;
	hepsE=(gamma1^r)*epsE;
	output.nevals=0;
	output.its=0;
%
%%%%%%%%%%%%%%%%%%% First-order criticality step
%
	firstordstop=0;
	deltaC=delta;
	itsC=0;
%
%	Initial model computation
	[gC,HC,evsC] = firstordmodel(x,fun,fx,deltaC*epsC);
	outputs.nevals = output.nevals+evsC;
%
	lC=0;
	ngC=norm(gC);
	if ngC >= epsC
		firstordstop = 1;
		output.flagC = 0;
	end
%
	while ~firstordstop
		ngCold = ngC;
		if ngCold>0
			[gC,HC,evsC] = firstordmodel(x,fun,fx,gamma1*deltaC*ngCold);
			ngC = norm(ngC);
			output.nevals = output.nevals+evsC;
		end
%
		if ( ngC < hepsC) || (gamma1*ngCold < ngC)
			firstordstop=1;
			if ngC < hepsC
				output.flagC = -1;
			else
				output.flagC = 1;
			end
		else
			deltaC = gamma1*deltaC;
			if (deltaC <= 1e-15)
				firstordstop = 1;
			end
			lC = lC+1;
		end
%
	end
%
%%%%%%%%%%%%%%%%%%% Second-order criticality step
%
	secondordstop=0;
	deltaE=delta;
	itsE=0;
%
%	Initial model computation
	[gE,HE,evsE] = secondordmodel(x,fun,fx,deltaE*epsE);
	outputs.nevals = output.nevals+evsE;
%
	lE=0;
	ldaE=max(-min(eig(HE)),0);
	if ldaE >= epsE
		secondordstop = 1;
		output.flagE = 0;
	end
%
	while ~secondordstop
		ldaEold = ldaE;
		if ldaEold>0
			[gE,HE,evsE] = secondordmodel(x,fun,fx,gamma1*deltaE*ldaEold);
			ldaE = max(-min(eig(HE)),0);
			output.nevals = output.nevals+evsE;
		end
%
		if ( ldaE < hepsE) || (gamma1*ldaEold < ldaE)
			secondordstop=1;
			if ldaE < hepsE
				output.flagE = -1;
			else
				output.flagE = 1;
			end
		else
			deltaE = gamma1*deltaE;
			if (deltaE <= 1e-15)
				secondordstop = 1;
			end
			lE = lE+1;
		end
%
	end
%
%%%%%%%%%%%%%%%%%%% Selecting final value
%
	output.its = min(lC,lE);
	delta = max(deltaC,deltaE);
%
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [gC,HC,nevalsC] = firstordmodel(x,fun,fx,delta)
%
% Computation of a first-order derivative-free model of the function f around
% a given point.
%
% [GC,HC] = firstordermodel(X,FX,FUN) returns the gradient and Hessian of a 
% linear or quadratic model with a value FX at X and a gradient GC at X. The 
% model is built so as to be fully-linear.
%
% Inputs:
%   x : a vector of length n
%   fun : a function from R^n to R, supposed to be twice continuously
%   differentiable
%   fx : the value of the function FUN at X
%   delta : a radius for a trust-region in which the model must be
%   fully linear
%
% Outputs :
%   gC : the gradient of the model at x
%   HC : the Hessian of the model at x
%   nevalsC : the number of calls to the function fun while building the
%   model
%
% C. Royer - July 22, 2018
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
	n = length(x);
	A = eye(n);
	nevalsC = 0;
	gC = zeros(n,1);
	HC = zeros(n);
	valsC = zeros(n,2);
%
%   Quadratic approximation allowed
	linear = 0;
%
	for i=1:n
		fip = feval(fun,x+delta*A(:,i));
%		nevalsC = nevalsC + 1;
		fim = feval(fun,x-delta*A(:,i));
		nevalsC = nevalsC + 2;
%		gC(i) = (fip-fx)/delta;
		gci = (fip-fim)/(2*delta);
		if ~isinf(gci) || ~isnan(gci)
			gC(i) = gci;
		end
		if ~linear
			Hcii = (fip-2*fx+fim)/(delta^2);
			if ~isinf(Hcii) || ~isnan(Hcii)
				HC(i,i) = Hcii;
			else
				HC(i,i) = 1;
			end
		end
	end
%
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [gE,HE,nevalsE] = secondordmodel(x,fun,fx,delta)
%
% Computation of a second-order derivative-free model of the function f around
% a given point.
%
% [GE,HE] = secondordermodel(X,FX,FUN) returns the gradient and Hessian of a 
% quadratic model with a value FX at X, a gradient GE at X and a Hessian HE at
% X. The model is built so as to be fully quadratic.
%
% Inputs:
%   x : a vector of length n
%   fun : a function from R^n to R, supposed to be twice continuously
%   differentiable
%   fx : the value of the function FUN at X
%   delta : a radius for a trust-region in which the model must be
%   fully linear
%
% Outputs :
%   gE : the gradient of the model at x
%   HE : the Hessian of the model at x
%   nevalsE : the number of calls to the function fun while building the
%   model
%
% C. Royer - July 22, 2018
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
	n = length(x);
	A = eye(n);
	nevalsE = 0;
	gE = zeros(n,1);
	HE = zeros(n);
%	Normalization to be sure that all points are in the trust-region
	alpha = delta/sqrt(2);
%
%	Vectors of function values
	fpm = zeros(n,2);
%
	infind = zeros(n,1);
	for i=1:n
		fpm(i,1) = feval(fun,x+alpha*A(:,i));
		fpm(i,2) = feval(fun,x-alpha*A(:,i));
		nevalsE = nevalsE + 2;
		if isinf(fpm(i,1)) || isinf(fpm(i,2))
			infind(i) = 1;
		else
			gE(i) = (fpm(i,1)-fpm(i,2))/(2*alpha);
			HE(i,i) = (fpm(i,1)-2*fx+fpm(i,2))/(alpha^2);
		end
	end
%
	for i=1:n
		if infind(i)==0
			for j=(i+1):n
			    ft = feval(fun,x+alpha*A(:,i)+alpha*A(:,j));
			    nevalsE = nevalsE + 1;
			    if ~isinf(ft)
				    heij = (ft-fpm(i,1)-fpm(j,1)+fx)/(alpha^2);
					if ~isinf(heij) && ~isnan(heij)
						HE(i,j) = heij;
					end
				end
			end
		end
	end
%
	HE = 1/2*(HE+HE');
%
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
