function [xsol,fsol,exitflag,output] = destress(x0,fun,gfun,Hfun,delta0,...
tols,maxits,boundtr,dfo,ftarget,varargin)
%
%	A wrapper for calling the destress (DEcoupling Steps in a Trust-REgionS 
%	Strategy) method.
%
%	xsol = destress(x0,fun) applies the DESTRESS algorithm to minimize the 
%	function defined by fun starting at x0, and returns the best solution 
%	point obtained in xsol. The algorithm employed in the minimization 
%	process is a derivative-free version of the DESTRESS method.
%
%	xsol = destress(x0,fun,gfun,Hfun) applies the DESTRESS algorithm to 
%	minimize the function defined by fun starting at x0, and returns the 
%	best solution point obtained as xsol. The first and second-order 
%	derivatives are assumed to be provided through the function handles 
%	gfun and hfun, respectively. When no derivatives are available, 
%	gfun and hfun are set as [].
%
%	xsol = destress(x0,fun,gfun,Hfundelta0) applies the DESTRESS algorithm 
%	to minimize the function fun starting at x0 with the value delta0 for 
%	the initial trust-region radius (set delta0=[] for the default value).
%
%	xsol = destress(x0,fun,gfun,Hfun,delta0,tols) specifies the tolerances 
%	for the convergence criterion. If tols is a vector of length 2, the 
%	method terminates if a point x such that
%
%		norm(gfun(x)) < tols(1)	and	lambdamin(Hfun(x)) > -tols(2),
%
%	where lambdamin(.) represents the minimum eigenvalue.
%	If tols is a unique value, the same tolerance is used for both 
%	tolerances. (Set tols=[] for the default values.)
%
%	xsol = destress(x0,fun,gfun,Hfun,delta0,tols,maxits) specifies a 
%	maximum number of iterations allowed for the algorithm (set maxits=[] 
%	for the default value).
%
%	xsol = destress(x0,fun,gfun,Hfun,delta0,tols,maxits,boundtr) indicates 
%	whether a lower bound on the size of the trust-region should be 
%	enforced.
%
%	xsol = destress(x0,fun,gfun,Hfun,delta0,tols,maxits,boundtr,dfo) 
%	specifies whether the method should use derivative or not.
%	If either gfun or Hfun is empty, the method will choose to run in 
%	derivative-free mode automatically.
%
%	[xsol,fsol] = destress(x0,fun,...) returns the best solution computed, 
%	xsol, and the corresponding value fsol=fun(xsol).
%
%	[xsol,fsol,exitflag] = destress(x0,fun,...) additionally returns a 
%	flag related to the run of the method.
%
%	[xsol,fsol,exitflag,output] = destress(x0,fun,...) additionally 
%	outputs a structure containing detailed information about the 
%	algorithmic run.
%
%	----------------
%	Input parameters
%	----------------
%
%	fun: Objective function handle
%
%	gfun: Gradient function handle
%
%	Hfun: Hessian function handle
%
%	x0:	Initial point - Must be provided
%
%	delta0: Initial trust-region radius - Can be empty, 1 by default
%
%	tols: Vector containing the desired tolerance
%		If length(tols)=2, tols(1) is the tolerance for the gradient 
%		norm and tols(2) is the tolerance on the opposite of the minimum 
%		Hessian eigenvalue.
%		If length(tols)=1, both tolerances will use the value in tols.
%		If tols=[], the default value [1e-4 1e-3] will be used.
%
%	boundtr: Boolean indicating if a lower bound on the trust-region 
%	parameter.
%		0: No lower bound (Default value)
%		1: A lower bound based upon the tolerances is defined.
%
%	dfo: Boolean indicating whether derivatives of the function should be 
%	used
%		0: The method does not use the derivatives or derivatives are 
%		not provided (Default value when gfun=[] or Hfun=[])
%		1: The method uses the derivatives (Default value when gfun 
%		and Hfun are provided)
%
%	ftarget: Targeted value for the objective function
%		-Inf: No value pre-specified (Default value)
%		Any real value: The method stops whenever this value is reached.		
%
%	-----------------
%	Output parameters
%	-----------------
%
%	xsol,fsol: Best point xsol and its value fsol obtained by the method
%
%	exitflag: flag indicating what happened in the method
%		0: Method stopped because the tolerances were achieved
%		1: Method stopped with only one tolerance was achieved
%		2: Method stopped without achieving the tolerances, because the 
%		distance between two iterates has become less than
%			tolx = min(1e-12,min(tols))
%		3: Method stopped without achieving the tolerances, but because 
%		the maximal number of iterations has been reached.
%		-1: An error occurred and the method did not complete its run.
%
%	output:	structure containing the following information
%		nits: Number of iterations
%		ngsol: Gradient norm at xsol
%		lxsol: Minimum Hessian eigenvalue at xsol
%		deltasol: Final trust-region radius
%		nevals: Vector of length 3 containing the number of calls
%			nevals(1): to the function code fun
%			nevals(2): to the gradient code gfun
%			nevals(3): to the Hessian code Hfun
%		histF: History matrix of size (nits+1)*2
%			histF(:,1): function values at the current iterate
%			histF(:,2): number of function evaluations
%		histG: History matrix of size (nits+1)*2
%			histG(:,1): gradient norms at the current iterate
%			histG(:,2): number of gradient evaluations
%		histH: History matrix of size (nits+1)*2
%			histH(:,1): minimum (negative) Hessian eigenvalue at 
%			the current iterate
%			histH(:,2): number of Hessian evaluations
%
% C. W. Royer - October 9, 2018
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if (nargin<2)
	error('destress: At least two inputs required');
end
%
if isempty(x0)
	error('destress: Empty initial point, please input a initial vector');
else
	x0 = x0(:);
end
%
if isempty(fun)
	error('destress: Please input a valid function handle');
end
%
if (nargin < 5) || isempty(delta0)
	delta0 = 1;
elseif (delta0 <= 0)
	warning('destress: Nonpositive trust-region radius, replaced by 1');
	delta0 = 1;
end
%
if (nargin < 6) || isempty(tols)
	tols = [1e-4 1e-3];
elseif length(tols)==1;
	tols = tols*[1 1];
elseif length(tols)>=3
	error('destress: Wrong input for the tols parameter');
end
%
if (nargin < 7) || isempty(maxits)
	maxits = 2000*length(x0);
elseif (maxits <= 0)
	warning('destress: Nonpositive maximum of iterations, replaced by 2000n');
	maxits = 2000*length(x0);
end
%
if (nargin < 8) || isempty(boundtr)
	boundtr = 0;
elseif ~ismember(boundtr,[0 1])
	error('destress: Boundtr can only be 0 or 1.');
end
%
if (nargin < 9) || isempty(dfo) || ~ismember(dfo,[0 1])
	dfo = isempty(gfun) || isempty(Hfun);
	if dfo
		disp('destress will run in derivative-free mode');
		gfun = [];
		Hfun = [];
	else
		disp('destress will run in derivative-based mode');
	end
end
%
if (nargin < 10) || isempty(ftarget)
	ftarget = -Inf;
elseif isnan(ftarget) || ((ftarget~=-Inf) && ~isnumeric(ftarget))
	error('destress: Please input a real value or -Inf for ftarget.');
end
%
% Calling the destress method
[xsol,fsol,exitflag,output] = destressalgo(fun,gfun,Hfun,x0,delta0,tols,...
maxits,boundtr,dfo,ftarget);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
