function [xsol,fsol,flag,output] = destressalgo(fun,gfun,Hfun,x0,delta0,tols,...
maxits,boundtr,dfo,ftarget)
%
% An elementary implementation of DESTRESS (DEcoupling Steps in a Trust-REgionS
% Strategy), a second-order globally convergent trust-regionS method for 
% nonconvex optimization.
%
% XSOL = destressalgo(FUN,GFUN,HFUN,X0,DELTA0,TOLS,MAXITS,BOUNDTR,DFO,FTARGET) 
% yields the best point obtained by the method while attempting to satisfy 
% approximate optimality conditions within a bounded number of iterations.
%
%
%%%%%%%%%
% Inputs:
%%%%%%%%%
%	fun : a twice continuously differentiable function
%	gfun : the first derivative of fun - returns a vector value
%	Hfun : the second derivative of fun - returns a matrix value
%	x0 : a starting point for the method
%	delta0 : initial trust-region radius
%	tols : vector of length 2 containing the desired tolerances
%		tols(1) is the tolerance on the gradient norm
%		tols(2) is the tolerance on the minimum Hessian eigenvalue
%	maxits : maximum number of iterations
%	boundtr: boolean indicating whether a lower bound on the size of the 
%	trust regions should be enforced
%	dfo : boolean indicating whether the code is to be run with or without
%	derivatives
%	ftarget : targeted function value (derivative-free case only)
%
% Outputs:
%	xsol : final iterate
%	fsol : function value at xsol
%	flag : information flag
%		0 : Convergence has occurred
%		1 : Only one criterion is satisfied
%		2 : Distance between the iterates has become too short
%		3 : The maximal number of iterations has been reached
%		-1 : An error occurred.
%   output : record structure with relevant information
%       output.nits : number of iterations
%       output.ngsol : gradient norm at xsol (or first-order gradient model 
%		norm in case of a derivative-free model)
%       output.lsol : minimum Hessian eigenvalue at xsol (or minimum eigenvalue 
%		of the second-order model Hessian in case of a derivative-free model)
%       output.deltasol : final trust-region radius
%       output.nevals : vector of length 3 containing the number of calls:
%           nevals(1) : to the function code fun
%           nevals(2) : to the gradient code gfun
%           nevals(3) : to the Hessian code Hfun
%       output.histF : history matrix of size (nits+1)*2
%           histF(:,1) : function values at the current iterate
%           histF(:,2) : number of function evaluations
%       output.histG : history matrix of size (nits+1)*2
%           histG(:,1) : function values at the current iterate
%           histG(:,2) : number of function evaluations
%       output.histH : history matrix of size (nits+1)*2
%           histH(:,1) : function values at the current iterate
%           histH(:,2) : number of function evaluations
%
% C. Royer - October 9, 2018
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%%%%%%%%%%%%%%%%%%% Initializations : algorithmic parameters
	n = length(x0);
	gamma1 = 0.5;
	gamma2 = 2;
	eta1 = 0.25;
	eta2 = 0.75;
	flag = -1;
	stopping = 0;
	varit = Inf;
	nits = 0;
	tolg = tols(1);
	tolH = tols(2);
	tolx = min(1e-12,min(tolg,tolH));
%	Additional tolerance on the trust-region radius
	if ~dfo
		told = 0;
	else
		told = 1e-6;
	end
%
	newit = 0; % Avoids unecessary derivative computations 
%
%%%%%%%%%%%%%%%%%%% Initial TR radii and iterate
	delta = delta0;
	x = x0;
	fx = feval(fun,x);
	nevals = [1 0 0];
	if ~dfo
		gC = feval(gfun,x);
		nevals(2) = nevals(2) + 1;
		HC = feval(Hfun,x);
		nevals(3) = nevals(3) + 1;
		gE = gC;
		HE = HC;
	else
%		Computation of two derivative-free models through a criticality step
		[gC,HC,gE,HE,delta,outputcrit] = dcritstep(x0,fun,fx,delta0,gamma1,...
		tolg,tolH);
		nevals(1) = nevals(1) + outputcrit.nevals;
		nits = nits + outputcrit.its;
	end
%%%%%%%%%%%%%%%%%%% Preparing main loop
%
	ngC = norm(gC);
%
	if isinf(ngC) || isnan(ngC)
		stopping=1;
	end
%
%	Eigenvalue/eigenvector computation.
%	NOTE: eigs might be preferable in presence of sparsity
%	[vEmin,lEmin] = eigs(HE,1,'sa');
%
	if (sum(isinf(HE))+sum(isnan(HE))>0)
		vEmin=zeros(n,1);
		lEmin = 0;
	else
		[VE,DE] = eig(HE);
		[lEmin,IEmin] = min(diag(DE));
		vEmin = VE(:,IEmin(1));
%
	end
	taux = max(0,-lEmin);
%
%
%%%%%%%%%
	tolg = tolg*max(1,ngC);
	if taux>0
		tolH = tolH*max(1,taux);
	end
%%%%%%%%%
	if dfo
		deltamax = 10;
	else
		deltamax = Inf;
%		Modified version
		gamma2 = 1.1;
	end
%
	histF = [fx nevals(1)];
	histG = [ngC nevals(2)];
	histH = [lEmin nevals(3)];
%
	if ~dfo
		stopping = stopping || ((ngC < tolg) && (taux < tolH));
	else
		stopping = stopping || (fx < ftarget);
	end 
%
%%%%%%%%%%%%%%%%%%% MAIN LOOP %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	while ~stopping
%
		nits = nits + 1;
		s = zeros(n,1);
		d = 0;
		flagspb = -1;
%
%%%%%%%%%%%%%%%%%%% First-order subproblem
%
		xC = x;
		fxC = fx;
		varitC = 0;
%
		if boundtr
%		Presented in experiments 2017/04/08
			deltaC = delta*max(ngC,tolg);
		else
			deltaC = delta*ngC;
		end
%
		if (deltaC > 1e-15)
			[sC,dC,flagC] = cauchystep(gC,HC,deltaC);
		else
			sC = zeros(n,1);
			dC = 0;
			flagC = 1;
		end
%
		if ~flagC
%
			xC = x + sC;
			fxC = feval(fun,x+sC);
			nevals(1) = nevals(1)+1;
			varitC = norm(sC)/(1+norm(x));
%
		elseif (flagC~=1)
%
			stopping = 1;
			warning(...
			'destress: Cauchy step computation failed (zero or NaN step).');
			flag = -1;
		end
%
%%%%%%%%%%%%%%%%%%% Second-order subproblem solver
%
		xE = x;
		fxE = fx;
		varitE = 0;
%
		if boundtr
			deltaE = delta*max(-lEmin,tolH);
		else
			deltaE = delta*max(-lEmin,0);
		end
%
		if (deltaE > 1e-15)
			[sE,dE,flagE] = eigenstep(gE,HE,lEmin,vEmin,deltaE);
		else
			sE = zeros(n,1);
			dE = 0;
			flagE = 1;
		end
%
		if ~flagE
%
			xE = x + sE;
			fxE = feval(fun,xE);
			nevals(1) = nevals(1)+1;
			varitE = norm(sE)/(1+norm(x));
%
		elseif (flagE~=1)
%
			stopping = 1;
			warning('destress: Eigenstep computation failed');
			flag = -1;
%
		end
%
%%%%%%%%%%%%%%%%%%% Choosing the best solution
%
		if (max(varitC,varitE) > 0)
%
			if (fxE < fxC)
%
				xS = xE;
				fxS = fxE;
				varitS = varitE;
%
			else
%
				xS = xC;
				fxS = fxC;
				varitS = varitC;
%
			end
%
%			Computing the decrease ratio
%		
			rho = min((fx-fxS)/dC,(fx-fxS)/dE);
%
			if (rho > eta1)
				varit = varitS;
				x = xS;
				fx = fxS;
%				Modified 17-03-31
%				if (rho > eta2)
				delta = min(deltamax,gamma2*delta);
%				end
%
				if ~dfo
					gC = feval(gfun,x);
					nevals(2) = nevals(2) + 1;
					HC = feval(Hfun,x);
					nevals(3) = nevals(3) + 1;
					gE = gC;
					HE = HC;
				else
					[gC,HC,gE,HE,delta,outputcrit] = dcritstep(x,fun,fx,...
					delta,gamma1,tolg,tolH);
					nevals(1) = nevals(1) + outputcrit.nevals;
					nits = nits + outputcrit.its;
				end
%
				ngC = norm(gC);
%
				if (sum(isinf(HE))+sum(isnan(HE))>0)
					vEmin = zeros(n,1);
					lEmin = 0;
				else
					[VE,DE] = eig(HE);
					[lEmin,IEmin] = min(diag(DE));
					vEmin = VE(:,IEmin(1));
				end
%
				if ~dfo
					stopping = (ngC < tolg) && (lEmin > -tolH);
				else
					stopping = (fx < ftarget);
				end
				stopping = stopping || (varit < tolx) || (nits >= maxits);
				newit = 1;
%
			else
				delta = gamma1*delta;
				stopping = (nits >= maxits) || (delta < told);
				newit = 0;
			end
%
		else
%
			delta = gamma1*delta;
			stopping = (nits >= maxits) || (delta < told);
			newit = 0;
%
		end
%
		histF = [histF;fx nevals(1)];
		histG = [histG;ngC nevals(2)];
		histH = [histH;lEmin nevals(3)];
%
	end
%
%%%%%%%%%%%%%%%%%%% END OF MAIN LOOP %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%%%%%%%%%%%%%%%%%%% Final settings
%
	xsol = x;
	fsol = fx;
	output.nevals = nevals;
	output.nits = nits;
	output.ngsol = ngC;
	output.lsol = lEmin;
	output.deltasol=delta;
	output.histF = histF;
	output.histG = histG;
	output.histH = histH;
%
	if (~dfo && (ngC < tolg) && (lEmin > -tolH)) || (dfo && (fx < ftarget))
		flag = 0;
	elseif (~dfo && ((ngC < tolg) || (lEmin > -tolH))) || ...
	(dfo && (delta < told))
		flag = 1;
	elseif (varit < tolx)
		flag = 2;
	elseif (nits >= maxits)
		flag = 3;
	end
%
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [s,d,flag] = cauchystep(g,H,delta)
%
% First-order trust-region subproblem solver.
%
% The step is chosen as the Cauchy step, i.e. the minimum in the direction of
% the steepest descent.
%
% S = cauchystep(G,H,V,DELTA) returns said step for the subproblem
%   min G'S + 0.5*S'*H*S    s.t.    ||S|| <= DELTA.
%
% Inputs:
%   g : a vector
%   H : a matrix
%   delta : a trust-region radius
%
% Outputs:
%   s : the computed step - a vector
%   d : the decrease obtained, i.e. -g'*s-0.5*s'*H*s
%   flag : an indicator of the success of the algorithm
%       0 : the step was successfully computed and non-zero
%       1 : special case : the step was zero
%       -1 : an error occurred
%
% C. Royer - February 2016
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
	n = length(g);
	s = zeros(n,1);
	d = 0;
	flag = -1;
%
	if (delta <= 1e-15)
		error('TR subproblem : delta must be positive.')
	end
%
%%%%%%%%%%%%%%%%%%% Computation of the Cauchy step
%
	ng = norm(g);
%
	if (ng~=0)
		rayg = g'*H*g;
%
		if (rayg <= 0)
		    s = -(delta/ng)*g;
		else
		    t = (ng^2)/rayg;
		    s = -min(delta/ng,t)*g;
		end
	end
%
%	Final settings
%
	if (norm(s)~=0)
		flag = 0;
		d = -g'*s-0.5*s'*H*s;
	else
		flag = 1;
	end
%
	if isnan(d)
	    flag = -1;
	end
%
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [s,d,flag] = eigenstep(g,H,lmin,vmin,delta)
%
% Second-order trust-region subproblem solver.
%
% The step is chosen as the eigenstep, i.e. the minimum in the direction of
% the eigenvector corresponding to the minimum eigenvalue, provided this value
% is negative.
%
% S = eigenstep(G,H,LMIN,VMIN,DELTA) returns said step for the subproblem
%   min G'S + 0.5*S'*H*S    s.t.    ||S|| <= DELTA.
%
% Inputs:
%   g : a vector
%   H : a matrix
%   lmin : the minimum eigenvalue of H
%   vmin : an eigenvector of H associated with the eigenvalue L
%   delta : a trust-region radius
%
% Outputs:
%   s : the computed step - a vector
%   d : the decrease obtained, i.e. -g'*s-0.5*s'*H*s
%   flag : an indicator of the success of the algorithm
%       0 : the step was successfully computed and non-zero
%       1 : special case : the step was zero
%       -1 : an error occurred
%
% C. Royer - February 2016
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
	n = length(g);
	s = zeros(n,1);
	d = 0;
	flag = -1;
%
	if (delta <= 1e-15)
		error('TR subproblem : delta must be positive.')
	end
%
%%%%%%%%%%%%%%%%%%% Computation of the eigenstep in case of indefinite Hessian
%
	if (lmin < 0)
%
		if (g'*vmin > 0)
		    vmin = -vmin;
		end
%
		s = delta*(vmin/norm(vmin));
		d = -g'*s-0.5*s'*H*s;
%
	end
%
%	 Final settings
%
	if norm(s)~=0
		flag = 0;
	else
	    flag = 1;
	end
%
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

